"""
The :mod:`holisticai.datasets.plots` module includes bias plotters.
"""

from holisticai.datasets.plots._exploratory import (
    correlation_matrix_plot,
)

# All bias plotters
__all__ = [
    "correlation_matrix_plot",
]
